import QtQuick 2.7

import Qt3D.Core 2.0
import Qt3D.Render 2.0
import Qt3D.Input 2.0
import Qt3D.Extras 2.0
import QtQuick.Scene3D 2.0

Item {
    id: root
    anchors.fill: parent

    property var matrix
    property rect brect

    S2QMarker {
        id: tl
        color: "yellow"
        pos: Qt.vector2d(60,129)
        onMoved: recalc()
    }
    S2QMarker {
        id: tr
        color: "yellow"
        pos: Qt.vector2d(262,75)
        onMoved: recalc()
    }
    S2QMarker {
        id: bl
        color: "yellow"
        pos: Qt.vector2d(57,192) //was 292
        onMoved: recalc()
    }
    S2QMarker {
        id: br
        color: "yellow"
        pos: Qt.vector2d(265,230) // y was 278
        onMoved: recalc()
    }

    Square2Quad {
        id: square2Quad
    }

    Component.onCompleted: recalc()
    function recalc() {
        var a = square2Quad.preRectToQuad(0, 0, img.width, img.height,
                                          tl.cx, tl.cy,
                                          tr.cx, tr.cy,
                                          bl.cx, bl.cy,
                                          br.cx, br.cy);
        var bb = a[0];
        var ma = a[1];

        img.x = bb.x;
        img.y = bb.y;
        img.width = bb.width;
        img.height = bb.height;

        root.matrix = ma;
        root.brect = bb;
    }
    onMatrixChanged: {
        var bb = root.brect
        dots.clear();
        var src = [
                    [0, 0, "TopLeft"],
                    [bb.width, 0, "TopRight"],
                    [0, bb.height, "BottomLeft"],
                    [bb.width, bb.height, "BottomRight"],
                    [bb.width/2, bb.height/2, "C"]
                ];
        for(var i in src) {
            var v = Qt.vector4d(src[i][0], src[i][1], 0, 1);
            var r = norm(root.matrix.times(v));
            dots.append({dx: bb.x + r.x, dy: bb.y + r.y, txt: src[i][2]+"*"});
        }
        console.log("A");
    }

    function norm(v) {
        return Qt.vector3d(v.x/v.w, v.y/v.w, v.z/v.w);
    }

    Image {
        id: img
        source: "qrc:/assets/lavrov.jpeg"

        transformOrigin: Item.TopLeft
        transform: Matrix4x4 {
            matrix: root.matrix
        }
    }

    Flow {
        anchors.bottom: textArea.top
        anchors.left: textArea.left
        anchors.right: textArea.right
        anchors.bottomMargin: 5
        height: 45
        Link {
            text: "Identity"
            onClicked: {
                root.matrix = Qt.matrix4x4();
                root.brect = Qt.rect(0, 0, 100, 100);
            }
        }
        Text {
            text: " | "
        }
        Link {
            text: "Rotate Z 5º"
            property real a: 5 * Math.PI / 180
            onClicked: root.matrix = root.matrix.times(
                           Qt.matrix4x4(Math.cos(a), -Math.sin(a), 0, 0,
                                        Math.sin(a),  Math.cos(a), 0, 0,
                                        0,           0,            1, 0,
                                        0,           0,            0, 1))
        }
        Text {
            text: " | "
        }
        Link {
            text: "Translate +30"
            property real a: 30
            onClicked: root.matrix = root.matrix.times(
                           Qt.matrix4x4(1, 0, 0, a,
                                        0, 1, 0, a,
                                        0, 0, 1, 0,
                                        0, 0, 0, 1))
        }
        Text {
            text: " | "
        }
        Link {
            text: "Scale XY*1.2"
            property real a: 1.2
            onClicked: root.matrix = root.matrix.times(
                           Qt.matrix4x4(a, 0, 0, 0,
                                        0, a, 0, 0,
                                        0, 0, 1, 0,
                                        0, 0, 0, 1))
        }
        Text {
            text: " | "
        }
        Link {
            text: "Skew X 15º"
            property real a: Math.tan(15 * Math.PI / 180)
            onClicked: root.matrix = root.matrix.times(
                           Qt.matrix4x4(1, a, 0, 0,
                                        0, 1, 0, 0,
                                        0, 0, 1, 0,
                                        0, 0, 0, 1))
        }
        Text {
            text: " | (drag yellow markers to restore matrix)"
        }
    }


    Rectangle {
        id: textArea
        anchors.bottom: parent.bottom
        anchors.left: parent.left
        anchors.margins: 10
        color: "#CCCCCC"

        height: 150
        width: 250
        Text {
            anchors.fill: parent
            anchors.margins: 5
            text: "matrix M:\n"
                  + [root.matrix.m11.toFixed(3), root.matrix.m12.toFixed(3), root.matrix.m13.toFixed(3), root.matrix.m14.toFixed(3)].join(" ") + "\n"
            + [root.matrix.m21.toFixed(3), root.matrix.m22.toFixed(3), root.matrix.m23.toFixed(3), root.matrix.m24.toFixed(3)].join(" ") + "\n"
            + [root.matrix.m31.toFixed(3), root.matrix.m32.toFixed(3), root.matrix.m33.toFixed(3), root.matrix.m34.toFixed(3)].join(" ") + "\n"
            + [root.matrix.m41.toFixed(3), root.matrix.m42.toFixed(3), root.matrix.m43.toFixed(3), root.matrix.m44.toFixed(3)].join(" ") + "\n"
            + "\nbounding rect (around yellow markers):\n"
            + "x:"+root.brect.x+" y:"+root.brect.x+" w:"+root.brect.width+" y:"+root.brect.height;
        }

    }

    ListModel {
        id: dots
    }

    Repeater {
        model: dots
        S2QMarker {
            color: "red"
            width: 3
            pos: Qt.vector2d(dx,dy)
            text: txt
        }
    }

    Text {
        x: 40
        y: 20
        text: "Drag these yellow markers to generate new matrix M\nThis is a QtQuick 2D part and the image is morphed\nusing simple Item's 'transform' field and matrix M"
    }
    Text {
        x: 420
        y: 40
        text: "This is Qt3D part and the image is\ntransformed using Qt3D's Transform { matrix: M }"
    }
    Text {
        x: 420
        y: 75
        color: "red"
        font.bold: true
        text: ">>> This part behaves really weird <<<\nPress 'Identity' and 'Skew X 15º', the image won't move\nhowever all other transformations will have effect.\n'Rotate' will change scale?"
    }
    Text {
        x: 420
        y: 520
        text: "This is also Qt3D but the markers are positioned\nto manually calculated coordinates:\nr = m*s\nwhere s is [0, 0] to [1, 1]"
    }
}
